"""Fetcher for Mercurial repositories."""
import pathlib
import shutil
import subprocess  # noqa: S404
from typing import Optional

from yarl import URL

from cutty.repositories.domain.fetchers import fetcher
from cutty.repositories.domain.matchers import scheme
from cutty.repositories.domain.revisions import Revision


@fetcher(match=scheme("file", "http", "https", "ssh"))
def hgfetcher(
    url: URL, destination: pathlib.Path, revision: Optional[Revision]
) -> None:
    """Fetch the repository using hg."""
    executable = shutil.which("hg")

    def hg(
        *args: str, cwd: Optional[pathlib.Path] = None
    ) -> subprocess.CompletedProcess[str]:
        """Run a hg command."""
        if executable is None:
            raise RuntimeError("cannot locate hg")

        return subprocess.run(  # noqa: S603
            [executable, *args], check=True, capture_output=True, text=True, cwd=cwd
        )

    if destination.exists():
        hg("pull", cwd=destination)
    else:
        hg("clone", "--noupdate", str(url), str(destination))

    options = ["--rev", revision] if revision is not None else []
    hg("update", *options, cwd=destination)



"""Configuration for projects generated from Cookiecutter templates."""
import json
import pathlib
from dataclasses import dataclass
from typing import Optional
from typing import Sequence

from cutty.filestorage.domain.files import RegularFile
from cutty.filesystems.domain.purepath import PurePath
from cutty.templates.domain.bindings import Binding


PROJECT_CONFIG_FILE = "cutty.json"


@dataclass
class ProjectConfig:
    """Project configuration."""

    template: str
    bindings: Sequence[Binding]
    directory: Optional[pathlib.PurePosixPath] = None


def createprojectconfigfile(project: PurePath, config: ProjectConfig) -> RegularFile:
    """Create a JSON file with the settings and bindings for a project."""
    directory = str(config.directory) if config.directory is not None else None
    path = project / PROJECT_CONFIG_FILE
    data = {
        "template": {"location": config.template, "directory": directory},
        "bindings": {binding.name: binding.value for binding in config.bindings},
    }
    blob = json.dumps(data).encode()

    return RegularFile(path, blob)


def readprojectconfigfile(project: pathlib.Path) -> ProjectConfig:
    """Load the project configuration."""
    path = project / PROJECT_CONFIG_FILE
    text = path.read_text()
    data = json.loads(text)

    if not isinstance(data, dict):
        raise TypeError(f"{path}: project configuration must be 'dict', got {data!r}")

    template = data["template"]["location"]

    if not isinstance(template, str):
        raise TypeError(f"{path}: template location must be 'str', got {template!r}")

    directory = data["template"]["directory"]

    if not (directory is None or isinstance(directory, str)):
        raise TypeError(
            f"{path}: template directory must be 'str' or 'None', got {template!r}"
        )

    bindings = [Binding(key, value) for key, value in data["bindings"].items()]

    return ProjectConfig(
        template,
        bindings,
        directory=pathlib.PurePosixPath(directory) if directory is not None else None,
    )



"""Matching URLs and paths."""
import pathlib
from collections.abc import Callable

from yarl import URL


Matcher = Callable[[URL], bool]
PathMatcher = Callable[[pathlib.Path], bool]


def scheme(*args: str) -> Matcher:
    """Return a matcher for the given URL schemes."""
    schemes = set(args)

    def _match(url: URL) -> bool:
        return url.scheme in schemes or ("file" in schemes and not url.scheme)

    return _match



"""Create a project from a Cookiecutter template."""
import pathlib
from collections.abc import Sequence
from typing import Optional

import platformdirs
from lazysequence import lazysequence

from cutty.filestorage.adapters.cookiecutter import createcookiecutterstorage
from cutty.filesystems.domain.purepath import PurePath
from cutty.repositories.adapters.storage import getdefaultrepositoryprovider
from cutty.templates.adapters.cookiecutter.binders import bindcookiecuttervariables
from cutty.templates.adapters.cookiecutter.config import findcookiecutterhooks
from cutty.templates.adapters.cookiecutter.config import findcookiecutterpaths
from cutty.templates.adapters.cookiecutter.config import loadcookiecutterconfig
from cutty.templates.adapters.cookiecutter.projectconfig import createprojectconfigfile
from cutty.templates.adapters.cookiecutter.projectconfig import ProjectConfig
from cutty.templates.adapters.cookiecutter.render import createcookiecutterrenderer
from cutty.templates.domain.bindings import Binding
from cutty.templates.domain.renderfiles import renderfiles


def create(
    template: str,
    *,
    extrabindings: Sequence[Binding] = (),
    no_input: bool = False,
    checkout: Optional[str] = None,
    outputdir: Optional[pathlib.Path] = None,
    directory: Optional[pathlib.PurePosixPath] = None,
    overwrite_if_exists: bool = False,
    skip_if_file_exists: bool = False,
    outputdirisproject: bool = False,
    createrepository: bool = True,
    createconfigfile: bool = True,
) -> None:
    """Generate a project from a Cookiecutter template."""
    cachedir = pathlib.Path(platformdirs.user_cache_dir("cutty"))
    templatedir = getdefaultrepositoryprovider(cachedir)(template, revision=checkout)

    if directory is not None:
        templatedir = templatedir.joinpath(*directory.parts)

    if outputdir is None:
        outputdir = pathlib.Path.cwd()

    config = loadcookiecutterconfig(template, templatedir)
    render = createcookiecutterrenderer(templatedir, config)
    bindings = bindcookiecuttervariables(
        config.variables,
        render,
        interactive=not no_input,
        bindings=extrabindings,
    )

    projectfiles = lazysequence(
        renderfiles(findcookiecutterpaths(templatedir, config), render, bindings)
    )
    if not projectfiles:  # pragma: no cover
        return

    strip = 0 if directory is None else len(directory.parts)

    if outputdirisproject:
        projectdir = outputdir
        strip += 1
    else:
        projectdir = outputdir / projectfiles[0].path.parts[strip]

    hookfiles = lazysequence(
        renderfiles(findcookiecutterhooks(templatedir), render, bindings)
    )
    projectconfigfile = (
        createprojectconfigfile(
            PurePath(*projectdir.relative_to(outputdir).parts),
            ProjectConfig(template, bindings, directory=directory),
        )
        if createconfigfile
        else None
    )

    with createcookiecutterstorage(
        outputdir,
        projectdir,
        overwrite_if_exists,
        skip_if_file_exists,
        hookfiles,
        createrepository,
    ) as storage:
        for projectfile in projectfiles.release():
            if strip:
                path = PurePath(*projectfile.path.parts[strip:])
                projectfile = projectfile.withpath(path)

            storage.add(projectfile)

        if projectconfigfile is not None:
            storage.add(projectconfigfile)



"""Compatibility shim for contextlib."""
import contextlib
from collections.abc import Callable
from collections.abc import Iterator
from typing import TypeVar


T = TypeVar("T")


def contextmanager(
    func: Callable[..., Iterator[T]]
) -> Callable[..., contextlib.AbstractContextManager[T]]:
    """Fix annotations of functions decorated by contextlib.contextmanager."""
    result = contextlib.contextmanager(func)
    result.__annotations__ = {
        **func.__annotations__,
        "return": contextlib.AbstractContextManager[T],
    }
    return result



"""Rendering."""
from __future__ import annotations

import functools
from collections.abc import Callable
from collections.abc import Sequence
from typing import Any
from typing import Mapping
from typing import TypeVar

from cutty.filestorage.domain.files import Executable
from cutty.filestorage.domain.files import RegularFile
from cutty.filestorage.domain.files import SymbolicLink
from cutty.filesystems.domain.path import Path
from cutty.filesystems.domain.purepath import PurePath
from cutty.templates.domain.bindings import Binding
from cutty.templates.domain.variables import GenericVariable
from cutty.templates.domain.variables import Variable


T = TypeVar("T")

GenericRenderer = Callable[[T, Sequence[Binding]], T]
Renderer = GenericRenderer[Any]
GenericRenderContinuation = Callable[[T, Sequence[Binding], Renderer], T]
RenderContinuation = GenericRenderContinuation[Any]
RenderRegistry = Mapping[type[Any], RenderContinuation]


def bindrendercontinuation(
    rendercontinuation: GenericRenderContinuation[T], render: Renderer
) -> GenericRenderer[T]:
    """Bind the third argument of a render continuation."""

    def _wrapper(value: T, bindings: Sequence[Binding]) -> T:
        value = rendercontinuation(value, bindings, render)
        return value

    return _wrapper


def createrenderer(renderregistry: RenderRegistry) -> Renderer:
    """Create a renderer."""

    @functools.singledispatch
    def _dispatch(value: T, bindings: Sequence[Binding]) -> T:
        raise NotImplementedError(f"no renderer registered for {type(value)}")

    for rendertype, rendercontinuation in renderregistry.items():
        renderfunction = bindrendercontinuation(rendercontinuation, _dispatch)
        _dispatch.register(rendertype, renderfunction)

    return _dispatch


def asrendercontinuation(
    renderfunction: GenericRenderer[T],
) -> GenericRenderContinuation[T]:
    """Return a render continuation that ignores the third argument."""
    # Use Renderer instead of GenericRenderer[T], because passing a
    # generic render function does not work with mypy.
    # https://github.com/python/mypy/issues/1317

    def _wrapper(value: T, bindings: Sequence[Binding], render: Renderer) -> T:
        return renderfunction(value, bindings)

    return _wrapper


def rendervariable(
    variable: Variable,
    bindings: Sequence[Binding],
    render: Renderer,
) -> Variable:
    """Render a variable by rendering its default and choices."""
    return Variable(
        variable.name,
        variable.description,
        variable.type,
        render(variable.default, bindings),
        tuple(render(choice, bindings) for choice in variable.choices),
        variable.interactive,
    )


def renderpurepath(
    path: PurePath,
    bindings: Sequence[Binding],
    render: Renderer,
) -> PurePath:
    """Render a path by rendering its parts."""
    return PurePath(*(render(part, bindings) for part in path.parts))


def renderpath(
    path: Path,
    bindings: Sequence[Binding],
    render: Renderer,
) -> Path:
    """Render a path by rendering its parts."""
    return Path(
        *(render(part, bindings) for part in path.parts),
        filesystem=path.filesystem,
    )


def renderregularfile(
    file: RegularFile,
    bindings: Sequence[Binding],
    render: Renderer,
) -> RegularFile:
    """Render a file by rendering its path and contents."""
    cls = Executable if isinstance(file, Executable) else RegularFile
    path = render(file.path, bindings)
    text = file.blob.decode()
    text = render(text, bindings)
    return cls(path, text.encode())


def rendersymlink(
    symlink: SymbolicLink, bindings: Sequence[Binding], render: Renderer
) -> SymbolicLink:
    """Render a symbolic link by rendering its path and target."""
    return SymbolicLink(
        render(symlink.path, bindings),
        render(symlink.target, bindings),
    )


defaultrenderregistry: RenderRegistry = {
    GenericVariable: rendervariable,
    PurePath: renderpurepath,
    Path: renderpath,
    RegularFile: renderregularfile,
    SymbolicLink: rendersymlink,
}



"""File storage observers."""
from cutty.filestorage.domain.files import File
from cutty.filestorage.domain.storage import FileStorage


class FileStorageObserver:
    """Base class for file storage observers."""

    def begin(self) -> None:
        """A storage transaction was started."""

    def add(self, file: File) -> None:
        """A file was added to the transaction."""

    def commit(self) -> None:
        """A storage transaction was completed."""

    def rollback(self) -> None:
        """A storage transaction was aborted."""


def observe(storage: FileStorage, observer: FileStorageObserver) -> FileStorage:
    """Wrap a storage with an observer."""

    class _ObservableFileStorage(FileStorage):
        def begin(self) -> None:
            storage.begin()
            observer.begin()

        def add(self, file: File) -> None:
            storage.add(file)
            observer.add(file)

        def commit(self) -> None:
            storage.commit()
            observer.commit()

        def rollback(self) -> None:
            storage.rollback()
            observer.rollback()

    return _ObservableFileStorage()



"""Determining storage locations for URLs."""
import pathlib
from collections.abc import Callable

from yarl import URL


Store = Callable[[URL], pathlib.Path]


def defaultstore(url: URL) -> pathlib.Path:
    """Return the relative path to the repository within the storage location."""
    path = pathlib.PurePosixPath(url.path)
    path = path.relative_to(path.parent)
    return pathlib.Path(path)



"""Update a project with changes from its Cookiecutter template."""
import hashlib
import tempfile
from collections.abc import Iterator
from pathlib import Path
from pathlib import PurePosixPath
from typing import Optional
from typing import Sequence

import pygit2

from cutty.compat.contextlib import contextmanager
from cutty.filestorage.adapters.observers.git import commit
from cutty.filestorage.adapters.observers.git import LATEST_BRANCH
from cutty.filestorage.adapters.observers.git import UPDATE_BRANCH
from cutty.filestorage.adapters.observers.git import UPDATE_BRANCH_REF
from cutty.filestorage.adapters.observers.git import UPDATE_MESSAGE
from cutty.services.create import create
from cutty.templates.adapters.cookiecutter.projectconfig import readprojectconfigfile
from cutty.templates.domain.bindings import Binding


def checkoutemptytree(repositorypath: Path) -> None:
    """Check out an empty tree from the repository."""
    repository = pygit2.Repository(repositorypath)
    oid = repository.TreeBuilder().write()
    repository.checkout_tree(repository[oid])


@contextmanager
def createworktree(
    repositorypath: Path,
    branch: str,
    *,
    checkout: bool = True,
) -> Iterator[Path]:
    """Create a worktree for the branch in the repository."""
    repository = pygit2.Repository(repositorypath)

    with tempfile.TemporaryDirectory() as directory:
        name = hashlib.blake2b(branch.encode(), digest_size=32).hexdigest()
        path = Path(directory) / name
        worktree = repository.add_worktree(name, path, repository.branches[branch])

        if not checkout:
            # Emulate `--no-checkout` by checking out an empty tree after the fact.
            # https://github.com/libgit2/libgit2/issues/5949
            checkoutemptytree(path)

        yield path

    # Prune with `force=True` because libgit2 thinks `worktree.path` still exists.
    # https://github.com/libgit2/libgit2/issues/5280
    worktree.prune(True)


def cherrypick(repositorypath: Path, reference: str) -> None:
    """Cherry-pick the commit onto the current branch."""
    repository = pygit2.Repository(repositorypath)
    oid = repository.references[reference].resolve().target
    repository.cherrypick(oid)

    if repository.index.conflicts:
        paths = {
            side.path
            for _, ours, theirs in repository.index.conflicts
            for side in (ours, theirs)
            if side is not None
        }
        raise RuntimeError(f"Merge conflicts: {', '.join(paths)}")

    commit(repository, message=UPDATE_MESSAGE)
    repository.state_cleanup()


def createbranch(
    repositorypath: Path, branch: str, *, target: str, force: bool = False
) -> None:
    """Create a branch pointing to the given target, another branch."""
    repository = pygit2.Repository(repositorypath)
    commit = repository.branches[target].peel()
    repository.branches.create(branch, commit, force=force)


def updatebranch(repositorypath: Path, branch: str, *, target: str) -> None:
    """Update a branch to the given target, another branch."""
    repository = pygit2.Repository(repositorypath)
    commit = repository.branches[target].peel()
    repository.branches[branch].set_target(commit.id)


def update(
    *,
    projectdir: Optional[Path] = None,
    extrabindings: Sequence[Binding] = (),
    no_input: bool = False,
    checkout: Optional[str] = None,
    directory: Optional[PurePosixPath] = None,
) -> None:
    """Update a project with changes from its Cookiecutter template."""
    if projectdir is None:
        projectdir = Path.cwd()

    projectconfig = readprojectconfigfile(projectdir)
    extrabindings = list(projectconfig.bindings) + list(extrabindings)

    if directory is None:
        directory = projectconfig.directory

    createbranch(projectdir, UPDATE_BRANCH, target=LATEST_BRANCH, force=True)

    with createworktree(projectdir, UPDATE_BRANCH, checkout=False) as worktree:
        create(
            projectconfig.template,
            outputdir=worktree,
            outputdirisproject=True,
            extrabindings=extrabindings,
            no_input=no_input,
            checkout=checkout,
            directory=directory,
        )

    cherrypick(projectdir, UPDATE_BRANCH_REF)

    updatebranch(projectdir, LATEST_BRANCH, target=UPDATE_BRANCH)


def continueupdate(*, projectdir: Optional[Path] = None) -> None:
    """Continue an update after conflict resolution."""
    if projectdir is None:
        projectdir = Path.cwd()

    commit(pygit2.Repository(projectdir), message=UPDATE_MESSAGE)
    updatebranch(projectdir, LATEST_BRANCH, target=UPDATE_BRANCH)


def resetmerge(repositorypath: Path, parent: str, cherry: str) -> None:
    """Reset only files that were touched by a cherry-pick.

    This emulates `git reset --merge HEAD` by performing a hard reset on the
    files that were updated by the cherry-picked commit, and resetting the index
    to HEAD.
    """
    repository = pygit2.Repository(repositorypath)
    repository.index.read_tree(repository.head.peel().tree)
    repository.index.write()

    parenttree = repository.branches[parent].peel(pygit2.Tree)
    cherrytree = repository.branches[cherry].peel(pygit2.Tree)
    diff = cherrytree.diff_to_tree(parenttree)
    paths = [
        file.path for delta in diff.deltas for file in (delta.old_file, delta.new_file)
    ]

    repository.checkout(strategy=pygit2.GIT_CHECKOUT_FORCE, paths=paths)


def skipupdate(*, projectdir: Optional[Path] = None) -> None:
    """Skip an update with conflicts."""
    if projectdir is None:
        projectdir = Path.cwd()

    resetmerge(projectdir, parent=LATEST_BRANCH, cherry=UPDATE_BRANCH)
    updatebranch(projectdir, LATEST_BRANCH, target=UPDATE_BRANCH)


def abortupdate(*, projectdir: Optional[Path] = None) -> None:
    """Abort an update with conflicts."""
    if projectdir is None:
        projectdir = Path.cwd()

    resetmerge(projectdir, parent=LATEST_BRANCH, cherry=UPDATE_BRANCH)
    updatebranch(projectdir, UPDATE_BRANCH, target=LATEST_BRANCH)



"""Repository providers."""
import pathlib
from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from types import MappingProxyType
from typing import Optional
from typing import Protocol

from yarl import URL

from cutty.filesystems.adapters.disk import DiskFilesystem
from cutty.filesystems.domain.filesystem import Filesystem
from cutty.filesystems.domain.path import Path
from cutty.repositories.domain.fetchers import Fetcher
from cutty.repositories.domain.fetchers import FetchMode
from cutty.repositories.domain.locations import aspath
from cutty.repositories.domain.locations import asurl
from cutty.repositories.domain.locations import Location
from cutty.repositories.domain.locations import parselocation
from cutty.repositories.domain.matchers import Matcher
from cutty.repositories.domain.matchers import PathMatcher
from cutty.repositories.domain.mounters import Mounter
from cutty.repositories.domain.revisions import Revision
from cutty.repositories.domain.stores import Store


class RepositoryProvider(Protocol):
    """The repository provider turns a repository URL into a filesystem path."""

    def __call__(
        self,
        location: str,
        revision: Optional[Revision] = None,
        fetchmode: FetchMode = FetchMode.ALWAYS,
    ) -> Path:
        """Return a path to the repository located at the given URL."""


Provider = Callable[[Location, Optional[Revision]], Optional[Filesystem]]


def provide(
    providers: Iterable[Provider], location: Location, revision: Optional[Revision]
) -> Path:
    """Provide a filesystem path for the repository located at the given URL."""
    for provider in providers:
        filesystem = provider(location, revision)
        if filesystem is not None:
            return Path(filesystem=filesystem)

    raise RuntimeError(f"unknown location {location}")


ProviderFactory = Callable[[Store, FetchMode], Provider]


def localprovider(*, match: PathMatcher, mount: Mounter) -> Provider:
    """Create a view onto the local filesystem."""

    def _localprovider(
        location: Location, revision: Optional[Revision]
    ) -> Optional[Filesystem]:
        try:
            path = location if isinstance(location, pathlib.Path) else aspath(location)
        except ValueError:
            return None
        else:
            return mount(path, revision) if match(path) else None

    return _localprovider


def _defaultmount(path: pathlib.Path, revision: Optional[Revision]) -> Filesystem:
    return DiskFilesystem(path)


def remoteproviderfactory(
    *,
    match: Optional[Matcher] = None,
    fetch: Iterable[Fetcher],
    mount: Optional[Mounter] = None,
) -> ProviderFactory:
    """Remote providers fetch the repository into local storage first."""
    fetch = tuple(fetch)
    _mount = mount if mount is not None else _defaultmount

    def _remoteproviderfactory(store: Store, fetchmode: FetchMode) -> Provider:
        def _remoteprovider(
            location: Location, revision: Optional[Revision]
        ) -> Optional[Filesystem]:
            url = location if isinstance(location, URL) else asurl(location)
            if match is None or match(url):
                for fetcher in fetch:
                    path = fetcher(url, store, revision, fetchmode)
                    if path is not None:
                        return _mount(path, revision)

            return None

        return _remoteprovider

    return _remoteproviderfactory


ProviderName = str
ProviderStore = Callable[[ProviderName], Store]
ProviderRegistry = Mapping[ProviderName, ProviderFactory]


_emptyproviderregistry: ProviderRegistry = MappingProxyType({})


def registerproviderfactory(
    providerregistry: ProviderRegistry,
    providername: ProviderName,
    providerfactory: ProviderFactory,
) -> ProviderRegistry:
    """Register a provider factory."""
    return {**providerregistry, providername: providerfactory}


def registerproviderfactories(
    providerregistry: ProviderRegistry = _emptyproviderregistry,
    **providerfactories: ProviderFactory,
) -> ProviderRegistry:
    """Register provider factories."""
    return {**providerregistry, **providerfactories}


def constproviderfactory(provider: Provider) -> ProviderFactory:
    """Create a provider factory that returns the given provider."""

    def _providerfactory(store: Store, fetchmode: FetchMode) -> Provider:
        return provider

    return _providerfactory


def registerprovider(
    providerregistry: ProviderRegistry,
    providername: ProviderName,
    provider: Provider,
) -> ProviderRegistry:
    """Register a provider factory."""
    return {**providerregistry, providername: constproviderfactory(provider)}


def registerproviders(
    providerregistry: ProviderRegistry = _emptyproviderregistry,
    **providers: Provider,
) -> ProviderRegistry:
    """Register provider factories."""
    providerfactories = {
        providername: constproviderfactory(provider)
        for providername, provider in providers.items()
    }
    return {**providerregistry, **providerfactories}


def _createprovider(
    providername: ProviderName,
    providerfactory: ProviderFactory,
    providerstore: ProviderStore,
    fetchmode: FetchMode,
) -> Provider:
    """Create a provider."""
    store = providerstore(providername)
    return providerfactory(store, fetchmode)


def _createproviders(
    providerregistry: ProviderRegistry,
    providerstore: ProviderStore,
    fetchmode: FetchMode,
    providername: Optional[ProviderName],
) -> Iterator[Provider]:
    """Create providers."""
    if providername is not None:
        providerfactory = providerregistry[providername]
        yield _createprovider(providername, providerfactory, providerstore, fetchmode)
    else:
        for providername, providerfactory in providerregistry.items():
            yield _createprovider(
                providername, providerfactory, providerstore, fetchmode
            )


def _splitprovidername(location: Location) -> tuple[Optional[ProviderName], Location]:
    """Split off the provider name from the URL scheme, if any."""
    if isinstance(location, URL):
        providername, _, scheme = location.scheme.rpartition("+")

        if providername:
            return providername, location.with_scheme(scheme)

    return None, location


def repositoryprovider(
    providerregistry: ProviderRegistry, providerstore: ProviderStore
) -> RepositoryProvider:
    """Return a repository provider."""

    def _provide(
        location: str,
        revision: Optional[Revision] = None,
        fetchmode: FetchMode = FetchMode.ALWAYS,
    ) -> Path:
        location_ = parselocation(location)
        providername, location_ = _splitprovidername(location_)
        providers = _createproviders(
            providerregistry, providerstore, fetchmode, providername
        )

        return provide(providers, location_, revision)

    return _provide



"""Jinja extensions."""
import json
import string
from secrets import choice
from typing import Any
from typing import Union

import jinja2.ext
import jinja2_time
from slugify import slugify


def jsonify(obj: Any) -> str:
    """Convert a Python object to JSON."""
    return json.dumps(obj, sort_keys=True, indent=4)


def random_ascii_string(length: int, punctuation: bool = False) -> str:
    """Create a random string."""
    corpus = string.ascii_letters
    if punctuation:
        corpus += string.punctuation
    return "".join(choice(corpus) for _ in range(length))


class JsonifyExtension(jinja2.ext.Extension):
    """Jinja extension to convert a Python object to JSON."""

    def __init__(self, environment: jinja2.Environment) -> None:
        """Initialize."""
        super().__init__(environment)
        environment.filters["jsonify"] = jsonify


class RandomStringExtension(jinja2.ext.Extension):
    """Jinja extension to create a random string."""

    def __init__(self, environment: jinja2.Environment) -> None:
        """Initialize."""
        super().__init__(environment)
        environment.globals["random_ascii_string"] = random_ascii_string


class SlugifyExtension(jinja2.ext.Extension):
    """Jinja extension to slugify a string."""

    def __init__(self, environment: jinja2.Environment) -> None:
        """Initialize."""
        super().__init__(environment)
        environment.filters["slugify"] = slugify


DEFAULT_EXTENSIONS: list[Union[str, type[jinja2.ext.Extension]]] = [
    JsonifyExtension,
    RandomStringExtension,
    SlugifyExtension,
    jinja2_time.TimeExtension,
]



"""Fetch a repository via HTTP."""
from pathlib import Path
from typing import Optional

import httpx
from yarl import URL

from cutty.repositories.domain.fetchers import fetcher
from cutty.repositories.domain.matchers import scheme
from cutty.repositories.domain.revisions import Revision


@fetcher(match=scheme("http", "https"))
def httpfetcher(url: URL, destination: Path, revision: Optional[Revision]) -> None:
    """Fetch via HTTP."""
    with httpx.stream("GET", str(url)) as response:
        response.raise_for_status()

        with destination.open(mode="wb") as io:
            for data in response.iter_bytes():
                io.write(data)



"""Decorator to disable runtime type-checking for a function."""
from typing import TYPE_CHECKING
from typing import TypeVar

if TYPE_CHECKING:  # pragma: no cover
    F = TypeVar("F")

    def typeguard_ignore(f: F) -> F:
        """This decorator is a noop during static type-checking."""
        return f


else:
    from typing import no_type_check as typeguard_ignore


__all__ = ["typeguard_ignore"]



"""Utilities for URLs and paths."""
import os.path
import pathlib
import platform
from typing import Union

from yarl import URL


Location = Union[pathlib.Path, URL]


def aspurewindowspath(url: URL) -> pathlib.PureWindowsPath:
    """Convert URL to Windows filesystem path."""
    if any(
        (
            url.scheme and url.scheme != "file",
            url.user,
            url.password,
            url.port,
            url.query_string,
            url.fragment,
        )
    ):
        raise ValueError(f"not a path: {url}")

    if url.host:
        return pathlib.PureWindowsPath(f"//{url.host}{url.path}")

    path = pathlib.PureWindowsPath(url.path[1:])
    if path.drive:
        return path

    return pathlib.PureWindowsPath(url.path)


def aspureposixpath(url: URL) -> pathlib.PurePosixPath:
    """Convert URL to POSIX filesystem path."""
    if any(
        (
            url.scheme and url.scheme != "file",
            url.host,
            url.user,
            url.password,
            url.port,
            url.query_string,
            url.fragment,
        )
    ):
        raise ValueError(f"not a path: {url}")

    return pathlib.PurePosixPath(url.path)


def aspath(url: URL) -> pathlib.Path:
    """Convert URL to filesystem path."""
    return pathlib.Path(
        aspurewindowspath(url)
        if platform.system() == "Windows"
        else aspureposixpath(url)
    )


def realpath(path: pathlib.Path) -> pathlib.Path:
    """Return the canonical path of the specified filename."""
    return pathlib.Path(os.path.realpath(path))


def asurl(path: pathlib.Path) -> URL:
    """Convert filesystem path to URL."""
    # Get an absolute path but avoid Path.resolve.
    # https://bugs.python.org/issue38671
    path = realpath(path)

    return URL(path.as_uri())


def parselocation(location: str) -> Location:
    """Construct a path or URL from a string."""
    path = pathlib.Path(location)

    try:
        exists = path.exists()
    except OSError:  # pragma: no cover
        exists = False  # illegal filename on Windows

    if not exists:
        url = URL(location)

        # A location without a scheme can be one of three things: a vanilla
        # path, a UNC-style Windows path, or a network-path reference (RFC 3986)
        # a.k.a. protocol-relative URL. The first two are paths, and the third
        # makes no sense without a base URI. So we treat all as paths.

        if url.scheme:
            return url

    return path



"""Binding variables in Cookiecutter templates."""
from collections.abc import Sequence

from cutty.templates.adapters.cookiecutter.prompts import prompt
from cutty.templates.domain.binders import binddefault
from cutty.templates.domain.binders import Binder
from cutty.templates.domain.binders import override
from cutty.templates.domain.binders import renderbindwith
from cutty.templates.domain.bindings import Binding
from cutty.templates.domain.render import Renderer
from cutty.templates.domain.variables import Variable


def bindcookiecuttervariables(
    variables: Sequence[Variable],
    render: Renderer,
    *,
    interactive: bool,
    bindings: Sequence[Binding],
) -> Sequence[Binding]:
    """Bind the template variables."""
    binder: Binder = prompt if interactive else binddefault
    binder = override(binder, bindings)
    return renderbindwith(binder)(render, variables)



"""Dictionary-based filesystem for your pocket."""
from collections.abc import Iterator
from typing import Any
from typing import cast

from cutty.filesystems.domain.filesystem import Access
from cutty.filesystems.domain.nodefs import FilesystemNode
from cutty.filesystems.domain.nodefs import NodeFilesystem
from cutty.filesystems.domain.purepath import PurePath


class DictFilesystemNode(FilesystemNode):
    """A node in a dict filesystem."""

    def __init__(self, node: Any) -> None:
        """Initialize."""
        self.node = node

    def is_dir(self) -> bool:
        """Return True if the node is a directory."""
        return isinstance(self.node, dict)

    def is_file(self) -> bool:
        """Return True if the node is a regular file."""
        return isinstance(self.node, str)

    def is_symlink(self) -> bool:
        """Return True if the node is a symbolic link."""
        return isinstance(self.node, PurePath)

    def read_bytes(self) -> bytes:
        """Return the file contents."""
        return self.read_text().encode()

    def read_text(self) -> str:
        """Return the file contents."""
        return cast(str, self.node)

    def readlink(self) -> PurePath:
        """Return the link target."""
        return cast(PurePath, self.node)

    def iterdir(self) -> Iterator[str]:
        """Iterate over the directory entries."""
        node: dict[str, Any] = self.node
        return iter(node.keys())

    def __truediv__(self, entry: str) -> FilesystemNode:
        """Return the given directory entry."""
        try:
            return DictFilesystemNode(self.node[entry])
        except KeyError:
            raise FileNotFoundError()

    def access(self, mode: Access) -> bool:
        """Return True if the user can access the node."""
        return Access.EXECUTE not in mode or self.is_dir()


class DictFilesystem(NodeFilesystem):
    """Dictionary-based filesystem for your pocket."""

    def __init__(self, tree: dict[str, Any]) -> None:
        """Initialize."""
        self.root = DictFilesystemNode(tree)



"""Disk-based filesystem implementation using pathlib."""
import os
import pathlib
from collections.abc import Iterator

from cutty.filesystems.domain.filesystem import Access
from cutty.filesystems.domain.filesystem import Filesystem
from cutty.filesystems.domain.purepath import PurePath


def _fromaccess(access: Access) -> int:
    mapping = {
        Access.READ: os.R_OK,
        Access.WRITE: os.W_OK,
        Access.EXECUTE: os.X_OK,
    }
    return sum(
        [mapping[flag] for flag in Access if flag and flag in access],
        start=os.F_OK,
    )


class DiskFilesystem(Filesystem):
    """Disk filesystem."""

    def __init__(self, root: pathlib.Path) -> None:
        """Inititalize."""
        self._root = root.resolve(strict=True)

    def resolve(self, path: PurePath) -> pathlib.Path:
        """Resolve the given path."""
        resolved = self._root.joinpath(*path.parts).resolve()
        resolved.relative_to(self._root)
        return resolved

    def is_dir(self, path: PurePath) -> bool:
        """Return True if this is a directory."""
        return self.resolve(path).is_dir()

    def iterdir(self, path: PurePath) -> Iterator[str]:
        """Iterate over the files in this directory."""
        for child in self.resolve(path).iterdir():
            child = child.relative_to(self._root)
            yield child.name

    def is_file(self, path: PurePath) -> bool:
        """Return True if this is a regular file (or a symlink to one)."""
        return self.resolve(path).is_file()

    def read_bytes(self, path: PurePath) -> bytes:
        """Return the contents of this file."""
        return self.resolve(path).read_bytes()

    def read_text(self, path: PurePath) -> str:
        """Return the contents of this file."""
        return self.resolve(path).read_text()

    def is_symlink(self, path: PurePath) -> bool:
        """Return True if this is a symbolic link."""
        resolved = self.resolve(path.parent) / path.name
        return resolved.is_symlink()

    def readlink(self, path: PurePath) -> PurePath:
        """Return the target of a symbolic link."""
        resolved = self.resolve(path.parent) / path.name
        target = resolved.readlink()
        return PurePath(*target.parts)

    def access(self, path: PurePath, mode: Access) -> bool:
        """Return True if the user can access the path."""
        return os.access(self.resolve(path), _fromaccess(mode))

    def eq(self, path: PurePath, other: PurePath) -> bool:
        """Return True if the paths are considered equal."""
        return pathlib.Path(*path.parts) == pathlib.Path(*other.parts)

    def lt(self, path: PurePath, other: PurePath) -> bool:
        """Return True if the path is less than the other path."""
        return pathlib.Path(*path.parts) < pathlib.Path(*other.parts)



"""Storing files in a Git repository."""
import contextlib
import os
import pathlib
from typing import Optional

import pygit2

from cutty.filestorage.domain.observers import FileStorageObserver


LATEST_BRANCH = "cutty/latest"
LATEST_BRANCH_REF = f"refs/heads/{LATEST_BRANCH}"
UPDATE_BRANCH = "cutty/update"
UPDATE_BRANCH_REF = f"refs/heads/{UPDATE_BRANCH}"
CREATE_MESSAGE = "Initial import"
UPDATE_MESSAGE = "Update project template"


def default_signature(repository: pygit2.Repository) -> pygit2.Signature:
    """Return the default signature."""
    with contextlib.suppress(KeyError):
        return pygit2.Signature(
            os.environ["GIT_AUTHOR_NAME"],
            os.environ["GIT_AUTHOR_EMAIL"],
        )
    return repository.default_signature  # pragma: no cover


def commit(
    repository: pygit2.Repository,
    *,
    message: str,
    signature: Optional[pygit2.Signature] = None,
) -> None:
    """Commit all changes in the repository.

    If there are no changes relative to the parent, this is a noop.
    """
    repository.index.add_all()

    tree = repository.index.write_tree()
    if not repository.head_is_unborn and tree == repository.head.peel().tree.id:
        return

    repository.index.write()

    if signature is None:
        signature = default_signature(repository)

    parents = [] if repository.head_is_unborn else [repository.head.target]
    repository.create_commit("HEAD", signature, signature, message, tree, parents)


class GitRepositoryObserver(FileStorageObserver):
    """Storage observer creating a git repository."""

    def __init__(self, *, project: pathlib.Path) -> None:
        """Initialize."""
        self.project = project

    def commit(self) -> None:
        """A storage transaction was completed."""
        try:
            repository = pygit2.Repository(self.project)
        except pygit2.GitError:
            repository = pygit2.init_repository(self.project)

        if UPDATE_BRANCH in repository.branches:
            # HEAD must point to update branch if it exists.
            head = repository.references["HEAD"].target
            if head != UPDATE_BRANCH_REF:
                raise RuntimeError(f"unexpected HEAD: {head}")

        message = (
            CREATE_MESSAGE
            if LATEST_BRANCH not in repository.branches
            else UPDATE_MESSAGE
        )

        commit(repository, message=message)

        if LATEST_BRANCH not in repository.branches:
            repository.branches.create(LATEST_BRANCH, repository.head.peel())



"""Command-line interface for updating projects from Cookiecutter templates."""
import pathlib
from typing import Optional

import click

from cutty.entrypoints.cli.create import extra_context_callback
from cutty.services.update import abortupdate
from cutty.services.update import continueupdate
from cutty.services.update import skipupdate
from cutty.services.update import update as service_update
from cutty.templates.domain.bindings import Binding


@click.argument("extra-context", nargs=-1, callback=extra_context_callback)
@click.option(
    "--no-input",
    is_flag=True,
    default=False,
    help="Do not prompt for template variables.",
)
@click.option(
    "-C",
    "--cwd",
    metavar="DIR",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, path_type=pathlib.Path
    ),
    help="Directory of the generated project.",
)
@click.option(
    "-c",
    "--checkout",
    metavar="REV",
    help="Branch, tag, or commit hash of the template repository.",
)
@click.option(
    "--directory",
    metavar="DIR",
    type=click.Path(path_type=pathlib.Path),
    help=(
        "Directory within the template repository that contains the "
        "cookiecutter.json file."
    ),
)
@click.option(
    "continue_",
    "--continue",
    is_flag=True,
    default=False,
    help="Resume updating after conflict resolution.",
)
@click.option(
    "--skip",
    is_flag=True,
    default=False,
    help="Skip the current update.",
)
@click.option(
    "--abort",
    is_flag=True,
    default=False,
    help="Abort the current update.",
)
def update(
    extra_context: dict[str, str],
    no_input: bool,
    cwd: Optional[pathlib.Path],
    checkout: Optional[str],
    directory: Optional[pathlib.Path],
    continue_: bool,
    skip: bool,
    abort: bool,
) -> None:
    """Update a project with changes from its template."""
    if continue_:
        continueupdate(projectdir=cwd)
        return

    if skip:
        skipupdate(projectdir=cwd)
        return

    if abort:
        abortupdate(projectdir=cwd)
        return

    extrabindings = [Binding(key, value) for key, value in extra_context.items()]
    service_update(
        extrabindings=extrabindings,
        no_input=no_input,
        projectdir=cwd,
        checkout=checkout,
        directory=pathlib.PurePosixPath(directory) if directory is not None else None,
    )



"""Filesystem abstraction."""
import abc
import enum
from collections.abc import Iterator

from cutty.filesystems.domain.purepath import PurePath


class Access(enum.Flag):
    """File access mode."""

    DEFAULT = 0
    EXECUTE = enum.auto()
    WRITE = enum.auto()
    READ = enum.auto()


class Filesystem(abc.ABC):
    """A filesystem abstraction."""

    @abc.abstractmethod
    def is_dir(self, path: PurePath) -> bool:
        """Return True if this is a directory."""

    @abc.abstractmethod
    def iterdir(self, path: PurePath) -> Iterator[str]:
        """Iterate over the files in this directory."""

    @abc.abstractmethod
    def is_file(self, path: PurePath) -> bool:
        """Return True if this is a regular file (or a symlink to one)."""

    @abc.abstractmethod
    def read_bytes(self, path: PurePath) -> bytes:
        """Return the contents of this file."""

    @abc.abstractmethod
    def read_text(self, path: PurePath) -> str:
        """Return the contents of this file."""

    @abc.abstractmethod
    def is_symlink(self, path: PurePath) -> bool:
        """Return True if this is a symbolic link."""

    @abc.abstractmethod
    def readlink(self, path: PurePath) -> PurePath:
        """Return the target of a symbolic link."""

    @abc.abstractmethod
    def access(self, path: PurePath, mode: Access) -> bool:
        """Return True if the user can access the path."""

    def eq(self, path: PurePath, other: PurePath) -> bool:
        """Return True if the paths are considered equal."""
        return path.parts == other.parts

    def lt(self, path: PurePath, other: PurePath) -> bool:
        """Return True if the path is less than the other path."""
        return path.parts < other.parts



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cutty',
 'cutty.compat',
 'cutty.entrypoints',
 'cutty.entrypoints.cli',
 'cutty.filestorage',
 'cutty.filestorage.adapters',
 'cutty.filestorage.adapters.observers',
 'cutty.filestorage.domain',
 'cutty.filesystems',
 'cutty.filesystems.adapters',
 'cutty.filesystems.domain',
 'cutty.repositories',
 'cutty.repositories.adapters',
 'cutty.repositories.adapters.fetchers',
 'cutty.repositories.adapters.providers',
 'cutty.repositories.domain',
 'cutty.services',
 'cutty.templates',
 'cutty.templates.adapters',
 'cutty.templates.adapters.cookiecutter',
 'cutty.templates.domain',
 'cutty.util']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.3,<4.0.0',
 'binaryornot>=0.4.4,<0.5.0',
 'click>=8.0.1,<9.0.0',
 'httpx>=0.18.1,<0.19.0',
 'jinja2-time>=0.2.0,<0.3.0',
 'lazysequence>=0.2.1,<0.4.0',
 'platformdirs>=2.0.2,<3.0.0',
 'pluggy>=0.13.1,<0.14.0',
 'pygit2>=1.5.0,<2.0.0',
 'python-slugify>=5.0.0,<6.0.0',
 'yarl>=1.6.3,<2.0.0']

entry_points = \
{'console_scripts': ['cutty = cutty.__main__:main']}

setup_kwargs = {
    'name': 'cutty',
    'version': '0.9.0',
    'description': 'cutty',
    'long_description': "cutty\n=====\n\n|PyPI| |Status| |Python Version| |License| |Read the Docs| |Tests| |Codecov|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/cutty.svg\n   :target: https://pypi.org/project/cutty/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/cutty.svg\n   :target: https://pypi.org/project/cutty/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/cutty\n   :target: https://pypi.org/project/cutty\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/pypi/l/cutty\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/cutty/latest.svg?label=Read%20the%20Docs\n   :target: https://cutty.readthedocs.io/\n   :alt: Read the documentation at https://cutty.readthedocs.io/\n.. |Tests| image:: https://github.com/cjolowicz/cutty/workflows/Tests/badge.svg\n   :target: https://github.com/cjolowicz/cutty/actions?workflow=Tests\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/cjolowicz/cutty/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/cjolowicz/cutty\n   :alt: Codecov\n\n\nAn experimental Cookiecutter clone :construction:\n\n\nRequirements\n------------\n\n* Python 3.9\n\n\nInstallation\n------------\n\nYou can install *cutty* via pip_ from PyPI_:\n\n.. code:: console\n\n   $ pip install cutty\n\n\nUsage\n-----\n\nPlease see the `Command-line Reference <Usage_>`_ for details.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*cutty* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\nThis project was generated from `@cjolowicz`_'s `Hypermodern Python Cookiecutter`_ template.\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/cjolowicz/cutty/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: https://cutty.readthedocs.io/en/latest/usage.html\n",
    'author': 'Claudio Jolowicz',
    'author_email': 'mail@claudiojolowicz.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cjolowicz/cutty',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Binding variables."""
from collections.abc import Callable
from collections.abc import Sequence
from typing import Any

from cutty.templates.domain.bindings import Binding
from cutty.templates.domain.render import Renderer
from cutty.templates.domain.variables import validate
from cutty.templates.domain.variables import Variable


Binder = Callable[[Variable], Binding]
RenderingBinder = Callable[[Renderer, Sequence[Variable]], Sequence[Binding]]


def bind(variable: Variable, value: Any) -> Binding:
    """Bind a variable to a value."""
    validate(value, variable)
    return Binding(variable.name, value)


def binddefault(variable: Variable) -> Binding:
    """Bind a variable to its default value."""
    return bind(variable, variable.default)


def override(binder: Binder, bindings: Sequence[Binding]) -> Binder:
    """Only use the binder if the binding does not yet exist."""
    if not bindings:
        return binder

    mapping = {binding.name: binding.value for binding in bindings}

    def _binder(variable: Variable) -> Binding:
        try:
            value = mapping[variable.name]
        except KeyError:
            return binder(variable)

        return bind(variable, value)

    return _binder


def renderbind(
    render: Renderer, binder: Binder, variables: Sequence[Variable]
) -> Sequence[Binding]:
    """Successively render and bind variables."""
    bindings: list[Binding] = []
    for variable in variables:
        variable = render(variable, bindings)
        binding = binder(variable)
        bindings.append(binding)
    return bindings


def renderbindwith(binder: Binder) -> RenderingBinder:
    """Render and bind variables using the given binder."""
    return lambda render, variables: renderbind(render, binder, variables)



"""Command-line interface."""
from cutty.entrypoints.cli._main import main
from cutty.entrypoints.cli.cookiecutter import cookiecutter
from cutty.entrypoints.cli.create import create
from cutty.entrypoints.cli.update import update


registercommand = main.command()

registercommand(create)
registercommand(update)
registercommand(cookiecutter)


__all__ = ["main"]



"""Fetch a repository via FTP."""
import pathlib
import shutil
import urllib.request
from typing import Optional

from yarl import URL

from cutty.repositories.domain.fetchers import fetcher
from cutty.repositories.domain.matchers import scheme
from cutty.repositories.domain.revisions import Revision


@fetcher(match=scheme("ftp"))
def ftpfetcher(
    url: URL, destination: pathlib.Path, revision: Optional[Revision]
) -> None:
    """Fetch via FTP."""
    # URLs with IPv6 literals are not supported, because
    # urllib.request.FTPHandler uses socket.gethostbyname.

    with urllib.request.urlopen(str(url)) as response:  # noqa: S310

        # FTPHandler does not set response.status, so we don't check it. Any
        # errors are raised as urllib.error.URLError.

        with destination.open(mode="wb") as io:
            shutil.copyfileobj(response, io)



"""Repository storage."""
from __future__ import annotations

import datetime
import hashlib
import json
import pathlib
import platform
import shutil
from collections.abc import Callable
from dataclasses import dataclass
from typing import Iterator
from typing import Optional

from yarl import URL

from cutty.repositories.adapters.registry import defaultproviderregistry
from cutty.repositories.domain.providers import ProviderName
from cutty.repositories.domain.providers import ProviderStore
from cutty.repositories.domain.providers import RepositoryProvider
from cutty.repositories.domain.providers import repositoryprovider
from cutty.repositories.domain.stores import Store


@dataclass
class StorageRecord:
    """Record describing the storage for a repository."""

    path: pathlib.Path
    url: URL
    provider: str
    updated: datetime.datetime

    @classmethod
    def load(cls, path: pathlib.Path) -> StorageRecord:
        """Deserialize a JSON file to a record."""
        text = (path / "config.json").read_text()
        data = json.loads(text)
        return cls(
            path,
            URL(data["url"]),
            data["provider"],
            datetime.datetime.fromisoformat(data["updated"]),
        )

    def dump(self) -> None:
        """Serialize the record to a JSON file."""
        data = {
            "url": str(self.url),
            "provider": self.provider,
            "updated": self.updated.isoformat(),
        }
        text = json.dumps(data)
        (self.path / "config.json").write_text(text)


# Windows does not properly support files and directories longer than
# 260 characters. Use a smaller digest size on this platform.
DIGEST_SIZE: int = 64 if platform.system() != "Windows" else 32


def hashurl(url: URL) -> str:
    """Return the hashsum for the given URL.

    Creates a digest using the BLAKE2b hash algorithm.
    """
    data = str(url).encode()
    return hashlib.blake2b(data, digest_size=DIGEST_SIZE).hexdigest()


Timer = Callable[[], datetime.datetime]


def defaulttimer() -> datetime.datetime:
    """Return the current time in UTC."""
    return datetime.datetime.now(tz=datetime.timezone.utc)


class RepositoryStorage:
    """Storage backend for repositories."""

    def __init__(self, path: pathlib.Path, *, timer: Timer = defaulttimer) -> None:
        """Initialize."""
        self.path = path
        self.path.mkdir(parents=True, exist_ok=True)
        self.timer = timer

    def _getrepositorypath(self, url: URL, *, provider: ProviderName) -> pathlib.Path:
        """Return the path to the repository."""
        h = hashurl(url)
        return self.path / "repositories" / provider / h[:2] / h[2:]

    def get(self, url: URL, *, provider: ProviderName) -> Optional[StorageRecord]:
        """Retrieve storage for a repository."""
        path = self._getrepositorypath(url, provider=provider)
        if not path.exists():
            return None

        record = StorageRecord.load(path)
        record.updated = self.timer()
        record.dump()

        return record

    def allocate(self, url: URL, *, provider: ProviderName) -> StorageRecord:
        """Allocate storage for a repository."""
        path = self._getrepositorypath(url, provider=provider)
        path.mkdir(parents=True)

        record = StorageRecord(path, url, provider, self.timer())
        record.dump()

        return record

    def list(self) -> Iterator[StorageRecord]:
        """Return the list of storage entries."""
        repositories = self.path / "repositories"
        if repositories.exists():
            for provider in repositories.iterdir():
                for prefix in provider.iterdir():
                    for path in prefix.iterdir():
                        yield StorageRecord.load(path)

    def clean(self, cutoff: datetime.datetime) -> Iterator[StorageRecord]:
        """Remove storage entries older than the given timestamp."""
        for record in self.list():
            if record.updated < cutoff:
                yield record
                shutil.rmtree(record.path)


def getdefaultproviderstore(
    path: pathlib.Path, *, timer: Timer = defaulttimer
) -> ProviderStore:
    """Return a provider store."""
    storage = RepositoryStorage(path, timer=timer)

    def providerstore(provider: str) -> Store:
        """Return a store function for the provider."""

        def store(url: URL) -> pathlib.Path:
            """Return a storage location for the URL."""
            record = storage.get(url, provider=provider)
            if record is None:
                record = storage.allocate(url, provider=provider)
            return record.path

        return store

    return providerstore


def getdefaultrepositoryprovider(
    path: pathlib.Path, *, timer: Timer = defaulttimer
) -> RepositoryProvider:
    """Return a repository provider."""
    return repositoryprovider(
        defaultproviderregistry,
        getdefaultproviderstore(path, timer=timer),
    )



"""Cookiecutter loader."""
import json
from collections.abc import Iterator
from typing import Any

from cutty.filesystems.domain.path import Path
from cutty.templates.domain.config import Config
from cutty.templates.domain.variables import Variable


def loadvalue(value: Any) -> Any:
    """Stringize scalars."""
    if isinstance(value, (bool, int, float)):
        return str(value)

    if isinstance(value, (str, dict)):
        return value

    raise RuntimeError(f"unsupported value type {type(value)}")  # pragma: no cover


def loadvariable(name: str, value: Any) -> Variable:
    """Load a variable."""
    if isinstance(value, list):
        choices = tuple(loadvalue(choice) for choice in value)
        [valuetype] = set(type(choice) for choice in choices)
        return Variable(
            name=name,
            description=name,
            type=valuetype,
            default=choices[0],
            choices=choices,
            interactive=True,
        )

    value = loadvalue(value)
    return Variable(
        name=name,
        description=name,
        type=type(value),
        default=value,
        choices=(),
        interactive=True,
    )


def loadcookiecutterconfig(template: str, path: Path) -> Config:
    """Load the configuration for a Cookiecutter template."""
    text = (path / "cookiecutter.json").read_text()
    data = json.loads(text)

    assert isinstance(data, dict) and all(  # noqa: S101
        isinstance(name, str) for name in data
    )

    data.setdefault("_template", template)

    settings = {name: value for name, value in data.items() if name.startswith("_")}

    variables = tuple(
        loadvariable(name, value)
        for name, value in data.items()
        if not name.startswith("_")
    )

    return Config(settings, variables)


def findcookiecutterpaths(path: Path, config: Config) -> Iterator[Path]:
    """Load project files in a Cookiecutter template."""
    for template_dir in path.iterdir():
        if all(token in template_dir.name for token in ("{{", "cookiecutter", "}}")):
            break
    else:
        raise RuntimeError("template directory not found")  # pragma: no cover

    yield template_dir


def findcookiecutterhooks(path: Path) -> Iterator[Path]:
    """Load hooks in a Cookiecutter template."""
    hooks = {"pre_gen_project", "post_gen_project"}
    hookdir = path / "hooks"
    if hookdir.is_dir():
        for path in hookdir.iterdir():
            if path.is_file() and not path.name.endswith("~") and path.stem in hooks:
                yield path
